"""Class implementation for the rotate_around_center interface.
"""

from typing import Union

import apysc as ap
from apysc._type.variable_name_interface import VariableNameInterface


class RotateAroundPointInterface(VariableNameInterface):

    def rotate_around_point(
            self, additional_rotation: Union[int, ap.Int],
            x: Union[int, ap.Int], y: Union[int, ap.Int]) -> None:
        """
        Add a rotation value around the specified point.

        Notes
        -----
        - This interface value is a relative value, not a absolute value,
            so if you call this value multiple times, total rotation will
            be cumulative.
        - This interface is not supported by the container instance, such
            as the `Sprite` class due to the HTML (SVG) specification.

        Parameters
        ----------
        additional_rotation : int or Int
            A value to add.
        x : int or Int
            Rotation point of a x-coordinate.
        y : int or Int
            Rotation point of a y-coordinate.

        References
        ----------
        - GraphicsBase rotate_around_point interface document
            - https://bit.ly/3z4Z4zg
        """
        with ap.DebugInfo(
                callable_=self.rotate_around_point, locals_=locals(),
                module_name=__name__, class_=RotateAroundPointInterface):
            from apysc._type import value_util
            from apysc._validation import number_validation
            number_validation.validate_integer(integer=additional_rotation)
            number_validation.validate_integer(integer=x)
            number_validation.validate_integer(integer=y)
            rotation_value_str: str = value_util.get_value_str_for_expression(
                value=additional_rotation)
            x_value_str: str = value_util.get_value_str_for_expression(value=x)
            y_value_str: str = value_util.get_value_str_for_expression(value=y)
            expression: str = (
                f'{self.variable_name}.rotate('
                f'{rotation_value_str}, {x_value_str}, {y_value_str});'
            )
            ap.append_js_expression(expression=expression)



"""Class implementation for the ellipse size interface.
"""

from typing import Dict

import apysc as ap
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface


class EllipseSizeInterface(VariableNameInterface, RevertInterface):

    _ellipse_size: ap.Int
    _ellipse_width: ap.Int
    _ellipse_height: ap.Int

    def _initialize_ellipse_size_if_not_initialized(self) -> None:
        """
        Initialize _ellipse_size attribute it hasn't been initialized yet.
        """
        if hasattr(self, '_ellipse_size'):
            return
        self._ellipse_size = ap.Int(0)

    @property
    def ellipse_size(self) -> ap.Int:
        """
        Get a ellipse size value.

        Returns
        -------
        ellipse_size : Int
            Ellipse size value.
        """
        with ap.DebugInfo(
                callable_='ellipse_size', locals_=locals(),
                module_name=__name__, class_=EllipseSizeInterface):
            from apysc._type import value_util
            self._initialize_ellipse_size_if_not_initialized()
            return value_util.get_copy(value=self._ellipse_size)

    @ellipse_size.setter
    def ellipse_size(self, value: ap.Int) -> None:
        """
        Update a ellipse size value. This inteface will updates
        both of the ellipse width and ellipse height attributes.

        Parameters
        ----------
        value : int or Int
            Ellipse size value.
        """
        with ap.DebugInfo(
                callable_='ellipse_size', locals_=locals(),
                module_name=__name__, class_=EllipseSizeInterface):
            from apysc._validation import number_validation
            if not isinstance(value, ap.Int):
                number_validation.validate_integer(integer=value)
                value = ap.Int(value)
            self._ellipse_size = value
            self._ellipse_width = value
            self._ellipse_height = value
            self._ellipse_size.\
                _append_incremental_calc_substitution_expression()
            self._ellipse_width.\
                _append_incremental_calc_substitution_expression()
            self._ellipse_height.\
                _append_incremental_calc_substitution_expression()
            self._append_ellipse_size_update_expression()

    def _append_ellipse_size_update_expression(self) -> None:
        """
        Append an ellipse size updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_ellipse_size_update_expression,
                locals_=locals(),
                module_name=__name__, class_=EllipseSizeInterface):
            from apysc._type import value_util
            self._initialize_ellipse_size_if_not_initialized()
            value_str: str = value_util.get_value_str_for_expression(
                value=self._ellipse_size)
            expression: str = (
                f'{self.variable_name}.radius({value_str});'
            )
            ap.append_js_expression(expression=expression)

    _ellipse_size_snapshots: Dict[str, int]

    def _make_snapshot(self, snapshot_name: str) -> None:
        """
        Make the value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not hasattr(self, '_ellipse_size_snapshots'):
            self._ellipse_size_snapshots = {}
        if self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._initialize_ellipse_size_if_not_initialized()
        self._ellipse_size_snapshots[snapshot_name] = int(
            self._ellipse_size._value)

    def _revert(self, snapshot_name: str) -> None:
        """
        Revert the value if the snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._ellipse_size._value = self._ellipse_size_snapshots[
            snapshot_name]



"""Implementations for the count of the current loop number
(the For or the other loop class nested number).
"""

import os


def get_current_loop_count() -> int:
    """
    Get the current loop count number.

    Returns
    -------
    loop_count : int
        Current loop count.
    """
    from apysc._expression import expression_file_util
    file_path: str = expression_file_util.LOOP_COUNT_FILE_PATH
    if not os.path.isfile(file_path):
        return 0
    with open(file_path, 'r') as f:
        file_str: str = f.read()
    if file_str == '':
        return 0
    loop_count: int = int(file_str)
    return loop_count


def increment_current_loop_count() -> None:
    """
    Increment the current loop count number.
    """
    from apysc._expression import expression_file_util
    file_path: str = expression_file_util.LOOP_COUNT_FILE_PATH
    current_loop_count: int = get_current_loop_count()
    with open(file_path, 'w') as f:
        f.write(str(current_loop_count + 1))


def decrement_current_loop_count() -> None:
    """
    Decrement the current loop count number.
    """
    from apysc._expression import expression_file_util
    file_path: str = expression_file_util.LOOP_COUNT_FILE_PATH
    current_loop_count: int = get_current_loop_count()
    current_loop_count -= 1
    current_loop_count = max(0, current_loop_count)
    with open(file_path, 'w') as f:
        f.write(str(current_loop_count))



"""Class implementation for the center y-coordinate interface.
"""

from typing import Dict

import apysc as ap
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface


class CyInterface(VariableNameInterface, RevertInterface):

    _cy: ap.Int

    def _initialize_cy_if_not_initialized(self) -> None:
        """
        Initialize _cy attribute it hasn't been initialized yet.
        """
        if hasattr(self, '_cy'):
            return
        self._cy = ap.Int(0)

    @property
    def y(self) -> ap.Int:
        """
        Get a center y-coordinate.

        Returns
        -------
        y : Int
            Center y-coordinate.
        """
        with ap.DebugInfo(
                callable_='y', locals_=locals(),
                module_name=__name__, class_=CyInterface):
            from apysc._type import value_util
            self._initialize_cy_if_not_initialized()
            return value_util.get_copy(value=self._cy)

    @y.setter
    def y(self, value: ap.Int) -> None:
        """
        Update a center y-coordinate.

        Parameters
        ----------
        value : int or Int
            Center y-coordinate value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='y', locals_=locals(),
                module_name=__name__, class_=CyInterface):
            from apysc._validation import number_validation
            number_validation.validate_integer(integer=value)
            if not isinstance(value, ap.Int):
                value = ap.Int(value)
            self._cy = value
            self._cy._append_incremental_calc_substitution_expression()
            self._append_cy_update_expression()

    def _append_cy_update_expression(self) -> None:
        """
        Append cy position updating expression.
        """
        with ap.DebugInfo(
                callable_=self._append_cy_update_expression, locals_=locals(),
                module_name=__name__, class_=CyInterface):
            from apysc._type import value_util
            self._initialize_cy_if_not_initialized()
            value_str: str = value_util.get_value_str_for_expression(
                value=self._cy)
            expression: str = (
                f'{self.variable_name}.cy({value_str});'
            )
            ap.append_js_expression(expression=expression)

    _cy_snapshots: Dict[str, int]

    def _make_snapshot(self, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not hasattr(self, '_cy_snapshots'):
            self._cy_snapshots = {}
        if self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._initialize_cy_if_not_initialized()
        self._cy_snapshots[snapshot_name] = int(self._cy._value)

    def _revert(self, snapshot_name: str) -> None:
        """
        Revert a value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._cy._value = self._cy_snapshots[snapshot_name]



"""Class implementation for mouse down interface.
"""

from typing import Any
from typing import Dict
from typing import Optional

from apysc._event.handler import Handler
from apysc._event.handler import HandlerData
from apysc._event.mouse_event_interface_base import MouseEventInterfaceBase


class MouseDownInterface(MouseEventInterfaceBase):

    _mouse_down_handlers: Dict[str, HandlerData]

    def mousedown(
            self, handler: Handler,
            options: Optional[Dict[str, Any]] = None) -> str:
        """
        Add mouse down event listener setting.

        Parameters
        ----------
        handler : Handler
            Callable that called when mouse is downed on this instance.
        options : dict or None, default None
            Optional arguments dictionary to be passed to handler.

        Returns
        -------
        name : str
            Handler's name.

        References
        ----------
        - Mousedown and mouseup interfaces document
            - https://bit.ly/3zgjOnF
        """
        import apysc as ap
        from apysc._validation.variable_name_validation import \
            validate_variable_name_interface_type
        with ap.DebugInfo(
                callable_=self.mousedown, locals_=locals(),
                module_name=__name__, class_=MouseDownInterface):
            from apysc._event.handler import append_handler_expression
            from apysc._event.handler import get_handler_name
            from apysc._type.variable_name_interface import \
                VariableNameInterface
            self_instance: VariableNameInterface = \
                validate_variable_name_interface_type(instance=self)
            self._initialize_mouse_down_handlers_if_not_initialized()
            name: str = get_handler_name(handler=handler, instance=self)
            self._set_mouse_event_handler_data(
                handler=handler, handlers_dict=self._mouse_down_handlers,
                options=options)
            self._append_mouse_event_binding_expression(
                name=name, mouse_event_type=ap.MouseEventType.MOUSEDOWN)
            e: ap.MouseEvent = ap.MouseEvent(this=self_instance)
            append_handler_expression(
                handler_data=self._mouse_down_handlers[name],
                handler_name=name, e=e)
            return name

    def _initialize_mouse_down_handlers_if_not_initialized(self) -> None:
        """
        Initialize _mouse_down_handlers attribute if it is not
        initialized yet.
        """
        if hasattr(self, '_mouse_down_handlers'):
            return
        self._mouse_down_handlers = {}

    def unbind_mousedown(self, handler: Handler) -> None:
        """
        Unbind specified handler's mouse down event.

        Parameters
        ----------
        handler : Handler
            Callable to be unbinded.

        References
        ----------
        - Mousedown and mouseup interfaces document
            - https://bit.ly/3zgjOnF
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.unbind_mousedown, locals_=locals(),
                module_name=__name__, class_=MouseDownInterface):
            self._initialize_mouse_down_handlers_if_not_initialized()
            self._unbind_mouse_event(
                handler=handler, mouse_event_type=ap.MouseEventType.MOUSEDOWN,
                handlers_dict=self._mouse_down_handlers)

    def unbind_mousedown_all(self) -> None:
        """
        Unbind all mouse down events.

        References
        ----------
        - Mousedown and mouseup interfaces document
            - https://bit.ly/3zgjOnF
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.unbind_mousedown_all, locals_=locals(),
                module_name=__name__, class_=MouseDownInterface):
            self._initialize_mouse_down_handlers_if_not_initialized()
            self._unbind_all_mouse_events(
                mouse_event_type=ap.MouseEventType.MOUSEDOWN,
                handlers_dict=self._mouse_down_handlers)



"""Class implementation for the ellipse height interface.
"""

from typing import Dict

import apysc as ap
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface


class EllipseHeightInterface(VariableNameInterface, RevertInterface):

    _ellipse_height: ap.Int

    def _initialize_ellipse_height_if_not_initialized(self) -> None:
        """
        Initialize _ellipse_height attribute it hasn't been initialized yet.
        """
        if hasattr(self, '_ellipse_height'):
            return
        self._ellipse_height = ap.Int(0)

    @property
    def ellipse_height(self) -> ap.Int:
        """
        Get ellipse height value.

        Returns
        -------
        ellipse_height : Int
            Ellipse height value.
        """
        with ap.DebugInfo(
                callable_='ellipse_height', locals_=locals(),
                module_name=__name__, class_=EllipseHeightInterface):
            from apysc._type import value_util
            self._initialize_ellipse_height_if_not_initialized()
            return value_util.get_copy(value=self._ellipse_height)

    @ellipse_height.setter
    def ellipse_height(self, value: ap.Int) -> None:
        """
        Update ellipse height value.

        Parameters
        ----------
        value : int or Int
            Ellipse height value.
        """
        with ap.DebugInfo(
                callable_='ellipse_height', locals_=locals(),
                module_name=__name__, class_=EllipseHeightInterface):
            from apysc._validation import number_validation
            number_validation.validate_integer(integer=value)
            if isinstance(value, int):
                value = ap.Int(value)
            self._ellipse_height = value
            self._ellipse_height.\
                _append_incremental_calc_substitution_expression()
            self._append_ellipse_height_update_expression()

    def _append_ellipse_height_update_expression(self) -> None:
        """
        Append ellipse height updating expression.
        """
        with ap.DebugInfo(
                callable_=self._append_ellipse_height_update_expression,
                locals_=locals(),
                module_name=__name__, class_=EllipseHeightInterface):
            from apysc._type import value_util
            self._initialize_ellipse_height_if_not_initialized()
            if hasattr(self, '_ellipse_width'):
                width_value_str: str = value_util.get_value_str_for_expression(
                    value=getattr(self, '_ellipse_width'))
            else:
                width_value_str = value_util.get_value_str_for_expression(
                    value=0)
            height_value_str: str = value_util.get_value_str_for_expression(
                value=self._ellipse_height)
            expression: str = (
                f'{self.variable_name}.radius({width_value_str}, '
                f'{height_value_str});'
            )
            ap.append_js_expression(expression=expression)

    _ellipse_height_snapshots: Dict[str, int]

    def _make_snapshot(self, snapshot_name: str) -> None:
        """
        Make the value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not hasattr(self, '_ellipse_height_snapshots'):
            self._ellipse_height_snapshots = {}
        if self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._initialize_ellipse_height_if_not_initialized()
        self._ellipse_height_snapshots[snapshot_name] = int(
            self._ellipse_height._value)

    def _revert(self, snapshot_name: str) -> None:
        """
        Revert the value if the snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._ellipse_height._value = self._ellipse_height_snapshots[
            snapshot_name]
